import copy
from typing import Any
import pyvesync.const as const
from defaults import TestDefaults
from pyvesync.device_map import DeviceMapTemplate
import call_json_switches
import call_json_outlets
import call_json_bulbs
import call_json_fans
import call_json_humidifiers
import call_json_purifiers


API_BASE_URL = const.API_BASE_URL
API_TIMEOUT = const.API_TIMEOUT
DEFAULT_TZ = const.DEFAULT_TZ
APP_VERSION = const.APP_VERSION
PHONE_BRAND = const.PHONE_BRAND
PHONE_OS = const.PHONE_OS
MOBILE_ID = const.MOBILE_ID
USER_TYPE = const.USER_TYPE

ALL_DEVICE_MAP_MODULES: list[DeviceMapTemplate] = [
    *call_json_bulbs.bulb_modules,
    *call_json_fans.fan_modules,
    *call_json_outlets.outlet_modules,
    *call_json_switches.switch_modules,
    *call_json_humidifiers.humidifier_modules,
    *call_json_purifiers.purifier_modules
]

ALL_DEVICE_MAP_DICT: dict[str, DeviceMapTemplate] = {
    module.setup_entry: module for module in ALL_DEVICE_MAP_MODULES
}
"""Contains dictionary of setup_entry: DeviceMapTemplate for all devices."""


"""
DEFAULT_BODY = Standard body for new device calls
DEFAULT_HEADER = standard header for most calls
DEFAULT_HEADER_BYPASS = standard header for most calls api V2
ENERGY_HISTORY = standard outlet energy history response
-------------------------------------------------------
login_call_body(email, pass) = body of login call
LOGIN_RET_BODY = return of login call
-------------------------------------------------------
get_devices_body() = body of call to get device list
LIST_CONF_10AEU = device list entry for 10A Europe outlet
LIST_CONF_10AUS = devlice list entry for 10A US outlet
LIST_CONF_7A = device list entry for 7A outlet
LIST_CONF_AIR = device list entry for air purifier
LIST_CONF_15A = device list entry for 15A outlet
LIST_CONF_WS = device list entry for wall switch
LIST_CONF_ESL100 = device list entry for bulb ESL100
LIST_CONF_OUTDOOR_1 = devlice list entry for outdoor outlet subDevice 1
LIST_CONF_OUTDOOR_2 = devlice list entry for outdoor outlet subDevice 2
DEVLIST_ALL = Return tuple for all devices
DEVLIST_10AEU = device list return for only 10A eu outlet
DEVLIST_10AUS = device list return for only 10A us outlet
DEVLIST_7A = device list return for only 7A outlet
DEVLIST_15A = device list return for only 15A outlet
DEVLIST_WS = device list return for only wall switch
DEVLIST_AIR = device list return for just air purifier
DEVLIST_ESL100 = device list return for just ESL100 bulb
DEVLIST_OUTDOOR_1 = device list return for outdoor outlet subDevice 1
DEVLIST_OUTDOOR_2 = device list return for outdoor outlet subDevice 2
---------------------------------------------------------
DETAILS_10A = Return for 10A outlet device details
DETAILS_15A = Return for 15A outlet device details
DETAILS_7A = Return for 7A outlet device details
DETAILS_WS = Return for wall switch device details
DETAILS_AIR = Return for Air Purifier device details
DETAILS_ESL100 = Return for ESL100 Bulb device details
DETAILS_OUTDOOR = return for 2 plug outdoor outlet

"""

BULBS = call_json_bulbs.BULBS
FANS = call_json_fans.FANS
OUTLETS = call_json_outlets.OUTLETS
SWITCHES = call_json_switches.SWITCHES


DEFAULT_HEADER = {
    'accept-language': 'en',
    'accountId': TestDefaults.account_id,
    'appVersion': APP_VERSION,
    'content-type': 'application/json',
    'tk': TestDefaults.token,
    'tz': DEFAULT_TZ,
}

DEFAULT_HEADER_BYPASS = {
    'Content-Type': 'application/json; charset=UTF-8',
    'User-Agent': 'okhttp/3.12.1'
}


def response_body(code: int = 0, msg: str | None = 'Success') -> dict[str, Any]:
    """Return base response dictionary."""
    return {
        'code': code,
        'traceId': TestDefaults.trace_id,
        'msg': msg
    }


class LoginResponses:
    GET_TOKEN_RESPONSE_SUCCESS = {
        "traceId": TestDefaults.trace_id,
        "code": 0,
        "msg": None,
        "result": {
            "accountID": TestDefaults.account_id,
            "avatarIcon": "",
            "nickName": "",
            "mailConfirmation": True,
            "registerSourceDetail": None,
            "verifyEmail": TestDefaults.email,
            "bizToken": None,
            "mfaMethodList": None,
            "authorizeCode": TestDefaults.authorization_code,
            "userType": "1",
        },
    }

    LOGIN_RESPONSE_SUCCESS = {
        "traceId": TestDefaults.trace_id,
        "code": 0,
        "msg": "request success",
        "module": None,
        "stacktrace": None,
        "result": {
            "birthday": "",
            "gender": "",
            "acceptLanguage": "en",
            "measureUnit": "Imperial",
            "weightG": 0.0,
            "heightCm": 0.0,
            "weightTargetSt": 0.0,
            "heightUnit": "FT",
            "heightFt": 0.0,
            "weightTargetKg": 0.0,
            "weightTargetLb": 0.0,
            "weightUnit": "LB",
            "maximalHeartRate": 0,
            "targetBfr": 0.0,
            "displayFlag": [],
            "targetStatus": 0,
            "realWeightKg": 0.0,
            "realWeightLb": 0.0,
            "realWeightUnit": "lb",
            "heartRateZones": 0.0,
            "runStepLongCm": 0.0,
            "walkStepLongCm": 0.0,
            "stepTarget": 0.0,
            "sleepTargetMins": 0.0,
            "regionType": 1,
            "currentRegion": TestDefaults.region,
            "token": TestDefaults.token,
            "countryCode": TestDefaults.country_code,
            "accountID": TestDefaults.account_id,
            "bizToken": None,
        },
    }

    LOGIN_RESPONSE_CROSS_REGION = {
        "traceId": "TRACE_ID",
        "code": -11260022,
        "msg": "login trigger cross region error.",
        "module": None,
        "stacktrace": None,
        "result": {
            "birthday": None,
            "gender": None,
            "acceptLanguage": None,
            "measureUnit": None,
            "weightG": None,
            "heightCm": None,
            "weightTargetSt": None,
            "heightUnit": None,
            "heightFt": None,
            "weightTargetKg": None,
            "weightTargetLb": None,
            "weightUnit": None,
            "maximalHeartRate": None,
            "targetBfr": None,
            "displayFlag": None,
            "targetStatus": None,
            "realWeightKg": None,
            "realWeightLb": None,
            "realWeightUnit": None,
            "heartRateZones": None,
            "runStepLongCm": None,
            "walkStepLongCm": None,
            "stepTarget": None,
            "sleepTargetMins": None,
            "regionType": 1,
            "currentRegion": TestDefaults.region,
            "token": None,
            "countryCode": TestDefaults.country_code,
            "accountID": TestDefaults.account_id,
            "bizToken": TestDefaults.biz_token,
        },
    }


class LoginRequests:
    authentication_success = {
        "acceptLanguage": "en",
        "accountID": "",
        "clientInfo": const.PHONE_BRAND,
        "clientType": const.CLIENT_TYPE,
        "clientVersion": const.APP_VERSION,
        "debugMode": False,
        "method": "authByPWDOrOTM",
        "osInfo": const.PHONE_OS,
        "terminalId": TestDefaults.terminal_id,
        "timeZone": const.DEFAULT_TZ,
        "token": "",
        "traceId": TestDefaults.trace_id,
        "userCountryCode": TestDefaults.country_code,
        "sourceAppID": TestDefaults.app_id,
        "appID": TestDefaults.app_id,
        "authProtocolType": "generic",
        "email": TestDefaults.email,
        "password": TestDefaults.password,
    }

    login_success = {
        "acceptLanguage": "en",
        "accountID": "",
        "clientInfo": const.PHONE_BRAND,
        "clientType": const.CLIENT_TYPE,
        "clientVersion": const.APP_VERSION,
        "debugMode": False,
        "method": "loginByAuthorizeCode4Vesync",
        "osInfo": const.PHONE_OS,
        "terminalId": TestDefaults.terminal_id,
        "timeZone": const.DEFAULT_TZ,
        "token": "",
        "traceId": TestDefaults.trace_id,
        "userCountryCode": TestDefaults.country_code,
        "authorizeCode": TestDefaults.authorization_code,
        "emailSubscriptions": False,
    }

    login_cross_region = {
        "acceptLanguage": "en",
        "accountID": "",
        "clientInfo": const.PHONE_BRAND,
        "clientType": const.CLIENT_TYPE,
        "clientVersion": const.APP_VERSION,
        "debugMode": False,
        "method": "loginByAuthorizeCode4Vesync",
        "osInfo": const.PHONE_OS,
        "terminalId": "2124569cc4905328aac52b39669b1b15e",
        "timeZone": const.DEFAULT_TZ,
        "token": "",
        "traceId": TestDefaults.trace_id,
        "userCountryCode": TestDefaults.country_code,
        "bizToken": TestDefaults.biz_token,
        "emailSubscriptions": False,
        "regionChange": "lastRegion",
    }


def BYPASS_V1_BODY(cid: str, config_module: str, json_cmd: dict):
    return {
        "traceId": TestDefaults.trace_id,
        "method": "bypass",
        "token": TestDefaults.token,
        "accountID": TestDefaults.account_id,
        "timeZone": DEFAULT_TZ,
        "acceptLanguage": "en",
        "appVersion": APP_VERSION,
        "phoneBrand": PHONE_BRAND,
        "phoneOS": PHONE_OS,
        "cid": cid,
        "configModule": config_module,
        "jsonCmd": json_cmd
    }


def login_call_body(email, password):
    json_object = {
        'acceptLanguage': 'en',
        'appVersion': APP_VERSION,
        'devToken': '',
        'email': email,
        'method': 'login',
        'password': password,
        'phoneBrand': PHONE_BRAND,
        'phoneOS': PHONE_OS,
        'timeZone': DEFAULT_TZ,
        'traceId': TestDefaults.trace_id,
        'userType': '1',
    }
    return json_object


def device_list_item(device_map: DeviceMapTemplate) -> dict:
    """Create a device list item from a device map.

    Parameters
    ----------
    device_map : DeviceMapTemplate
        Device map to create device list item from

    Returns
    -------
    dict
        Device list item dictionary
    """
    return {
        "deviceType": device_map.dev_types[0],
        "isOwner": True,
        "deviceName": TestDefaults.name(device_map.setup_entry),
        "deviceImg": "",
        "cid": TestDefaults.cid(device_map.setup_entry),
        "uuid": TestDefaults.uuid(device_map.setup_entry),
        "macID": TestDefaults.macid(device_map.setup_entry),
        "connectionType": "wifi",
        "type": device_map.product_type,
        "configModule": TestDefaults.config_module(device_map.setup_entry),
        "mode": None,
        "speed": None,
        "currentFirmVersion": None,
        "speed": None,
        "subDeviceNo": None,
        "extension": None,
        "deviceProps": None,
        "deviceStatus": const.DeviceStatus.ON,
        "connectionStatus": const.ConnectionStatus.ONLINE,
        "product_type": device_map.product_type
    }


class DeviceList:
    list_response_base: dict[str, Any] = {
        'code': 0,
        'traceId': TestDefaults.trace_id,
        'msg': 'Success',
        'result': {
            'pageNo': 1,
            'pageSize': 100,
            'total': 0,
            'list': [],
        }
    }
    device_list_base: dict[str, Any] = {
        'extension': None,
        'isOwner': True,
        'authKey': None,
        'deviceStatus': 'on',
        'connectionStatus': 'online',
        'connectionType': 'wifi',
        'mode': None,
        'speed': None,
        'deviceProps': None,
        'configModule': 'ConfigModule',
        'deviceRegion': TestDefaults.country_code,

    }

    bulbs = dict.fromkeys(call_json_bulbs.BULBS, "wifi-light")
    outlets = dict.fromkeys(call_json_outlets.OUTLETS, "wifi-switch")
    fans = dict.fromkeys(call_json_fans.FANS, "wifi-air")
    switches = dict.fromkeys(call_json_switches.SWITCHES, "Switches")
    # purifiers = dict.fromkeys(call_json_purifiers.)

    @classmethod
    def device_list_item(cls, module: DeviceMapTemplate):
        model_types = {**cls.bulbs, **cls.outlets, **cls.fans, **cls.switches}

        device_dict = cls.device_list_base
        model_dict = device_dict.copy()
        # model_dict['deviceType'] = setup_entry
        model_dict['deviceName'] = TestDefaults.name(module.setup_entry)
        model_dict['type'] = model_types.get(module.setup_entry)
        model_dict['cid'] = TestDefaults.cid(module.setup_entry)
        model_dict['uuid'] = TestDefaults.uuid(module.setup_entry)
        model_dict['macID'] = TestDefaults.macid(module.setup_entry)
        model_dict['deviceType'] = module.dev_types[0]
        if module.setup_entry == 'ESO15-TB':
            model_dict['subDeviceNo'] = 1
        return model_dict

    @classmethod
    def device_list_response(
        cls,
        setup_entrys: list[str] | str | None = None,
    ):
        """Class method that returns the api get_devices response

        Args:
            setup_entrys (list, str optional): List or string of setup_entry(s)
                to return. Defaults to None.
        """
        if setup_entrys is None:
            entry_list = ALL_DEVICE_MAP_DICT
        elif isinstance(setup_entrys, str):
            entry_list = {
                k: v for k, v in ALL_DEVICE_MAP_DICT.items() if k == setup_entrys
            }
        elif isinstance(setup_entrys, list):
            entry_list = {
                k: v for k, v in ALL_DEVICE_MAP_DICT.items() if k in setup_entrys
            }
        response_base = copy.deepcopy(cls.list_response_base)

        response_base['result']['list'] = []
        response_base['result']['total'] = 0
        for module in entry_list.values():
            response_base['result']['list'].append(cls.device_list_item(module))
            response_base['result']['total'] += 1
        return response_base

    LIST_CONF_7A = {
        'deviceType': 'wifi-switch-1.3',
        'extension': None,
        'macID': None,
        'type': 'wifi-switch',
        'deviceName': 'Name 7A Outlet',
        'connectionType': 'wifi',
        'uuid': None,
        'speed': None,
        'deviceStatus': 'on',
        'mode': None,
        'configModule': '7AOutlet',
        'currentFirmVersion': '1.95',
        'connectionStatus': 'online',
        'cid': '7A-CID',
    }

    LIST_CONF_15A = {
        'deviceType': 'ESW15-USA',
        'extension': None,
        'macID': None,
        'type': 'wifi-switch',
        'deviceName': 'Name 15A Outlet',
        'connectionType': 'wifi',
        'uuid': 'UUID',
        'speed': None,
        'deviceStatus': 'on',
        'mode': None,
        'configModule': '15AOutletNightlight',
        'currentFirmVersion': None,
        'connectionStatus': 'online',
        'cid': '15A-CID',
    }

    LIST_CONF_WS = {
        'deviceType': 'ESWL01',
        'extension': None,
        'macID': None,
        'type': 'Switches',
        'deviceName': 'Name Wall Switch',
        'connectionType': 'wifi',
        'uuid': 'UUID',
        'speed': None,
        'deviceStatus': 'on',
        'mode': None,
        'configModule': 'InwallswitchUS',
        'currentFirmVersion': None,
        'connectionStatus': 'online',
        'cid': 'WS-CID',
    }

    LIST_CONF_10AEU = {
        'deviceType': 'ESW01-EU',
        'extension': None,
        'macID': None,
        'type': 'wifi-switch',
        'deviceName': 'Name 10A Outlet',
        'connectionType': 'wifi',
        'uuid': 'UUID',
        'speed': None,
        'deviceStatus': 'on',
        'mode': None,
        'configModule': '10AOutletEU',
        'currentFirmVersion': None,
        'connectionStatus': 'online',
        'cid': '10A-CID',
    }

    LIST_CONF_10AUS = {
        'deviceType': 'ESW03-USA',
        'extension': None,
        'macID': None,
        'type': 'wifi-switch',
        'deviceName': 'Name 10A Outlet',
        'connectionType': 'wifi',
        'uuid': 'UUID',
        'speed': None,
        'deviceStatus': 'on',
        'mode': None,
        'configModule': '10AOutletUSA',
        'currentFirmVersion': None,
        'connectionStatus': 'online',
        'cid': '10A-CID',
    }

    LIST_CONF_OUTDOOR_1 = {
        'deviceRegion': 'US',
        'deviceName': 'Outdoor Socket B',
        'cid': 'OUTDOOR-CID',
        'deviceStatus': 'on',
        'connectionStatus': 'online',
        'connectionType': 'wifi',
        'deviceType': 'ESO15-TB',
        'type': 'wifi-switch',
        'uuid': 'UUID',
        'configModule': 'OutdoorSocket15A',
        'macID': None,
        'mode': None,
        'speed': None,
        'extension': None,
        'currentFirmVersion': None,
        'subDeviceNo': 1,
    }

    LIST_CONF_OUTDOOR_2 = {
        'deviceRegion': 'US',
        'deviceName': 'Outdoor Socket B',
        'cid': 'OUTDOOR-CID',
        'deviceStatus': 'on',
        'connectionStatus': 'online',
        'connectionType': 'wifi',
        'deviceType': 'ESO15-TB',
        'type': 'wifi-switch',
        'uuid': 'UUID',
        'configModule': 'OutdoorSocket15A',
        'macID': None,
        'mode': None,
        'speed': None,
        'extension': None,
        'currentFirmVersion': None,
        'subDeviceNo': 2,
    }

    LIST_CONF_ESL100 = {
        'deviceRegion': 'US',
        'deviceName': 'Etekcity Soft White Bulb',
        'cid': 'ESL100-CID',
        'deviceStatus': 'on',
        'connectionStatus': 'online',
        'connectionType': 'wifi',
        'deviceType': 'ESL100',
        'type': 'Wifi-light',
        'uuid': 'UUID',
        'configModule': 'WifiSmartBulb',
        'macID': None,
        'mode': None,
        'speed': None,
        'extension': None,
        'currentFirmVersion': None,
        'subDeviceNo': None,
    }

    LIST_CONF_ESL100CW = {
        'deviceRegion': 'US',
        'deviceName': 'ESL100CW NAME',
        'cid': 'ESL100CW-CID',
        'deviceStatus': 'on',
        'connectionStatus': 'online',
        'connectionType': 'wifi',
        'deviceType': 'ESL100CW',
        'type': 'Wifi-light',
        'uuid': 'ESL100CW-UUID',
        'configModule': 'WifiSmartBulb',
        'macID': None,
        'mode': None,
        'speed': None,
        'extension': None,
        'currentFirmVersion': None,
        'subDeviceNo': None,
    }

    LIST_CONF_AIR = {
        'deviceName': 'Name Air Purifier',
        'cid': 'AIRPUR-CID',
        'deviceStatus': 'on',
        'connectionStatus': 'online',
        'connectionType': 'wifi',
        'deviceType': 'LV-PUR131S',
        'type': 'wifi-air',
        'uuid': 'UUID',
        'configModule': 'AirPurifier131',
        'macID': None,
        'mode': 'manual',
        'speed': 'low',
        'extension': None,
        'currentFirmVersion': None,
    }

    LIST_CONF_DUAL200S = {
        'deviceRegion': 'EU',
        'isOwner': True,
        'authKey': None,
        'deviceName': '200S NAME',
        'cid': 'CID-200S',
        'deviceStatus': 'on',
        'connectionStatus': 'online',
        'connectionType': 'WiFi+BTOnboarding+BTNotify',
        'deviceType': 'LUH-D301S-WEU',
        'type': 'wifi-air',
        'uuid': 'UUID-200S',
        'configModule': 'WFON_AHM_LUH-D301S-WEU_EU',
        'macID': None,
        'mode': None,
        'speed': None,
        'extension': None,
        'currentFirmVersion': None,
        'subDeviceNo': None,
        'subDeviceType': None,
    }

    LIST_CONF_DIMMER = {
        "deviceRegion": "US",
        "deviceName": "Etekcity Dimmer Switch",
        "cid": "DIM-CID",
        "deviceStatus": "on",
        "connectionStatus": "online",
        "connectionType": "wifi",
        "deviceType": "ESWD16",
        "type": "Switches",
        "uuid": "DIM-UUID",
        "configModule": "WifiWallDimmer"
    }

    LIST_CONF_600S = {
        'deviceRegion': 'US',
        'deviceName': 'Bedroom Humidifier',
        'deviceImg': '',
        'cid': 'CID-600S',
        'deviceStatus': 'on',
        'connectionStatus': 'online',
        'connectionType': 'WiFi+BTOnboarding+BTNotify',
        'deviceType': 'LUH-A602S-WUS',
        'type': 'wifi-air',
        'uuid': 'UUID-600S',
        'configModule': 'WFON_AHM_LUH-A602S-WUS_US',
        'macID': None,
        'subDeviceNo': None,
        'subDeviceType': None,
        'deviceProp': None
    }

    LIST_CONF_ESL100MC = {
        "cid": "CID-ESL100MC",
        "uuid": "UUID-ESL100MC",
        "macID": None,
        "subDeviceNo": 0,
        "subDeviceType": None,
        "deviceName": "ESL100MC NAME",
        "configModule": "WiFi_Bulb_MulticolorBulb_US",
        "type": "Wifi-light",
        "deviceType": "ESL100MC",
        "deviceStatus": "on",
        "connectionType": "wifi",
        "currentFirmVersion": "1.0.12",
        "connectionStatus": "online",
        "speed": None,
        "extension": None,
        "deviceProp": None
    }

    LIST_CONF_LV131S = {
        'deviceName': 'LV131S NAME',
        'cid': 'CID-LV131S',
        'deviceStatus': 'on',
        'connectionStatus': 'online',
        'connectionType': 'wifi',
        'deviceType': 'LV-PUR131S',
        'type': 'wifi-air',
        'uuid': 'UUID-LV131S',
        'configModule': 'AirPurifier131',
        'macID': None,
        'mode': 'auto',
        'speed': None,
        'extension': None,
        'currentFirmVersion': None,
        'subDeviceNo': None,
        'subDeviceType': None
    }

    LIST_CONF_VALCENO = {
        "deviceName": "VALCENO NAME",
        "cid": "CID-VALCENO",
        "deviceStatus": "on",
        "connectionStatus": "online",
        "connectionType": "WiFi+BTOnboarding+BTNotify",
        "deviceType": "XYD0001",
        "type": "Wifi-light",
        "uuid": "UUID-VALCENO",
        "configModule": "VC_WFON_BLB_A19-MC_US",
        "macID": None,
        "subDeviceNo": None,
        "subDeviceType": None
    }

    API_URL = '/cloud/v1/deviceManaged/devices'

    METHOD = 'POST'

    FULL_DEV_LIST = [
        LIST_CONF_10AEU,
        LIST_CONF_10AUS,
        LIST_CONF_15A,
        LIST_CONF_7A,
        LIST_CONF_AIR,
        LIST_CONF_WS,
        LIST_CONF_ESL100,
        LIST_CONF_OUTDOOR_1,
        LIST_CONF_OUTDOOR_2,
        LIST_CONF_DIMMER,
        LIST_CONF_600S,
        LIST_CONF_LV131S,
        LIST_CONF_DUAL200S,
        LIST_CONF_ESL100CW,
        LIST_CONF_ESL100MC,
        LIST_CONF_VALCENO

    ]

    @classmethod
    def DEVICE_LIST_RETURN(cls, dev_conf: dict) -> tuple:
        """Test the fan."""
        return (
            {
                'code': 0,
                'result':
                    {
                        'list': [dev_conf]
                    }
            },
            200
        )

    FAN_TEST = ({'code': 0, 'result': {'list': [LIST_CONF_600S, LIST_CONF_LV131S,
                                                LIST_CONF_DUAL200S]}}, 200)

    DEVLIST_ALL = ({'code': 0, 'result': {'list': FULL_DEV_LIST}}, 200)

    DEVLIST_7A = ({'code': 0, 'result': {'list': [LIST_CONF_7A]}}, 200)

    DEVLIST_15A = ({'code': 0, 'result': {'list': [LIST_CONF_15A]}}, 200)

    DEVLIST_10AEU = ({'code': 0, 'result': {'list': [LIST_CONF_10AEU]}}, 200)

    DEVLIST_10AUS = ({'code': 0, 'result': {'list': [LIST_CONF_10AUS]}}, 200)

    DEVLIST_WS = ({'code': 0, 'result': {'list': [LIST_CONF_WS]}}, 200)

    DEVLIST_DIMMER = ({'code': 0, 'result': {'list': [LIST_CONF_DIMMER]}}, 200)

    DEVLIST_AIR = ({'code': 0, 'result': {'list': [LIST_CONF_AIR]}}, 200)

    DEVLIST_ESL100 = ({'code': 0, 'result': {'list': [LIST_CONF_ESL100]}}, 200)

    DEVLIST_DUAL200S = ({'code': 0, 'result': {'list': [LIST_CONF_DUAL200S]}}, 200)

    DEVLIST_OUTDOOR = (
        {'code': 0, 'result': {'list': [LIST_CONF_OUTDOOR_1, LIST_CONF_OUTDOOR_2]}},
        200,
    )


class DeviceDetails:
    """Responses for get_details() method for all devices.

    class attributes:
    outlets : dict
        Dictionary of outlet responses for each device type.
    switches : dict
        Dictionary of switch responses for each device type.
    bulbs : dict
        Dictionary of bulb responses for each device type.
    fans : dict
        Dictionary of humidifier & air pur responses for each device type.
    all_devices : dict
        Dictionary of all device responses for each device type.

    Example
    -------
    outlets = {'ESW01-EU': {'switches': [{'outlet': 0, 'switch': 'on'}]}}
    """

    outlets = call_json_outlets.DETAILS_RESPONSES
    switches = call_json_switches.DETAILS_RESPONSES
    fans = call_json_fans.DETAILS_RESPONSES
    bulbs = call_json_bulbs.DETAILS_RESPONSES
    all_devices = {
        'outlets': outlets,
        'switches': switches,
        'fans': fans,
        'bulbs': bulbs
        }


DETAILS_BADCODE = {
    "code": 1000,
    "deviceImg": "",
    "activeTime": 1,
    "energy": 1,
    "power": "1",
    "voltage": "1",
}


STATUS_BODY = {
    'accountID': TestDefaults.account_id,
    'token': TestDefaults.token,
    'uuid': 'UUID',
    'timeZone': DEFAULT_TZ,
}


def off_body():
    body = STATUS_BODY
    body['status'] = 'off'
    return body, 200


def on_body():
    body = STATUS_BODY
    body['status'] = 'on'
    return body, 200



"""Thermostat device classes."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from pyvesync.base_devices.thermostat_base import VeSyncThermostat
from pyvesync.const import (
    ThermostatConst,
    ThermostatEcoTypes,
    ThermostatFanModes,
    ThermostatHoldStatus,
    ThermostatWorkModes,
)
from pyvesync.models.thermostat_models import (
    ResultThermostatDetails,
    ThermostatMinorDetails,
)
from pyvesync.utils.device_mixins import BypassV2Mixin, process_bypassv2_result
from pyvesync.utils.helpers import Helpers, Validators

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import ThermostatMap
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel

_LOGGER = logging.getLogger(__name__)


class VeSyncAuraThermostat(BypassV2Mixin, VeSyncThermostat):
    """VeSync Aura Thermostat."""

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: ThermostatMap,
    ) -> None:
        """Initialize VeSync Aura Thermostat."""
        super().__init__(details, manager, feature_map)

    def _process_details(self, details: ResultThermostatDetails) -> None:
        """Internal method to process thermostat details."""
        if ResultThermostatDetails.supportMode is not None:
            self.supported_work_modes = [
                ThermostatWorkModes(mode) for mode in ResultThermostatDetails.supportMode
            ]
        self.state.work_mode = ThermostatConst.WorkMode(details.workMode)
        self.state.work_status = ThermostatConst.WorkStatus(details.workStatus)
        self.state.fan_status = ThermostatConst.FanStatus(details.fanStatus)
        self.state.fan_mode = ThermostatConst.FanMode(details.fanMode)
        self.state.temperature_unit = details.tempUnit
        self.state.temperature = details.temperature
        self.state.humidity = details.humidity
        self.state.heat_to_temp = details.heatToTemp
        self.state.cool_to_temp = details.coolToTemp
        self.state.deadband = details.deadband
        self.state.lock_status = details.lockStatus
        self.state.schedule_or_hold = ThermostatConst.ScheduleOrHoldOption(
            details.scheduleOrHold
        )
        self.state.hold_end_time = details.holdEndTime
        self.state.hold_option = ThermostatConst.HoldOption(details.holdOption)
        self.state.eco_type = ThermostatConst.EcoType(details.ecoType)
        self.state.alert_status = details.alertStatus
        self.state.routine_running_id = details.routineRunningId
        self.state.routines = details.routines

    async def get_details(self) -> None:
        """Get the details of the thermostat."""
        r_dict = await self.call_bypassv2_api('getTsStatus')
        r_model = process_bypassv2_result(
            self, _LOGGER, 'get_details', r_dict, ResultThermostatDetails
        )
        if r_model is None:
            return
        self._process_details(r_model)

    async def get_configuration(self) -> None:
        """Get configuration or 'minor details'."""
        r_dict = await self.call_bypassv2_api('getTsMinorInfo')
        result = process_bypassv2_result(
            self, _LOGGER, 'get_configuration', r_dict, ThermostatMinorDetails
        )
        if result is None:
            return
        self.state.configuration = result

    async def _set_hold_status_api(
        self,
        key: str | None = None,
        temp: float | None = None,
        hold_status: ThermostatHoldStatus | None = None,
    ) -> bool:
        """Internal function to call the 'setHoldStatus' API."""
        if key is None and temp is None and hold_status is None:
            _LOGGER.debug('At least one of key, temp, or hold_status must be provided.')
            return False
        payload_data: dict[str, str | float] = {}
        if hold_status is not None and hold_status.value == ThermostatHoldStatus.CANCEL:
            payload_data = {'holdStatus': ThermostatHoldStatus.CANCEL}
        else:
            if key is None or temp is None:
                _LOGGER.debug('Either key or temp must be provided.')
                return False
            payload_data = {key: temp, 'holdStatus': ThermostatHoldStatus.SET.value}
        r_dict = await self.call_bypassv2_api('setHoldStatus', data=payload_data)
        result = Helpers.process_dev_response(_LOGGER, 'setHoldStatus', self, r_dict)
        return bool(result)

    async def set_temp_point(self, temperature: float) -> bool:
        return await self._set_hold_status_api('setTempPoint', temperature)

    async def cancel_hold(self) -> bool:
        """Cancel hold."""
        return await self._set_hold_status_api(hold_status=ThermostatHoldStatus.CANCEL)

    async def set_cool_to_temp(self, temperature: float) -> bool:
        """Set cool to temperature."""
        return await self._set_hold_status_api('coolToTemp', temperature)

    async def set_heat_to_temp(self, temperature: float) -> bool:
        """Set heat to temperature."""
        return await self._set_hold_status_api('heatToTemp', temperature)

    async def set_mode(self, mode: ThermostatWorkModes) -> bool:
        """Set thermostat mode."""
        if mode not in self.supported_work_modes:
            _LOGGER.debug('Invalid mode: %s', mode)
            return False
        payload_data = {'tsMode': mode}
        r_dict = await self.call_bypassv2_api('setTsMode', data=payload_data)
        result = Helpers.process_dev_response(_LOGGER, 'setTsMode', self, r_dict)
        return bool(result)

    async def set_fan_mode(self, mode: ThermostatFanModes) -> bool:
        """Set thermostat fan mode."""
        if mode not in self.fan_modes:
            _LOGGER.debug('Invalid fan mode: %s', mode)
            return False
        payload_data = {'fanMode': mode.value}
        r_dict = await self.call_bypassv2_api('setFanMode', data=payload_data)
        result = Helpers.process_dev_response(_LOGGER, 'setFanMode', self, r_dict)
        return bool(result)

    async def toggle_lock(self, toggle: bool, pin: int | str | None = None) -> bool:
        """Toggle thermostat lock status."""
        if toggle is True and pin is None:
            _LOGGER.debug('PIN required to lock the thermostat.')
            return False
        payload_data: dict[str, bool | str] = {'lockStatus': toggle}
        if toggle is True:
            pin_int = int(pin) if isinstance(pin, str) else pin
            if not Validators.validate_range(pin_int, 0, 9999):
                _LOGGER.debug('PIN must be between 0 and 9999.')
                return False
            payload_data['lockPinCode'] = f'{pin_int:0>4}'
        r_dict = await self.call_bypassv2_api('setLockStatus', data=payload_data)
        result = Helpers.process_dev_response(_LOGGER, 'toggle_lock_status', self, r_dict)
        return bool(result)

    async def set_eco_type(self, eco_type: ThermostatEcoTypes) -> bool:
        """Set thermostat eco type."""
        if eco_type not in self.eco_types:
            _LOGGER.debug('Invalid eco type: %s', eco_type)
            return False
        payload_data = {'ecoType': eco_type.value}
        r_dict = await self.call_bypassv2_api('setECOType', data=payload_data)
        result = Helpers.process_dev_response(_LOGGER, 'setEcoType', self, r_dict)
        return bool(result)



"""Base Devices for Outlets."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from pyvesync.base_devices.vesyncbasedevice import DeviceState, VeSyncBaseToggleDevice
from pyvesync.const import (
    ENERGY_HISTORY_MAP,
    EnergyIntervals,
    NightlightModes,
    OutletFeatures,
)
from pyvesync.models.outlet_models import RequestEnergyHistory, ResponseEnergyHistory
from pyvesync.utils.helpers import Helpers

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import OutletMap
    from pyvesync.models.outlet_models import ResponseEnergyResult
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel

logger = logging.getLogger(__name__)


class OutletState(DeviceState):
    """Base state class for Outlets.

    This class holds all of the state information for the outlet devices. The state
    instance is stored in the `state` attribute of the outlet device. This is only
    for holding state information and does not contain any methods for controlling
    the device or retrieving information from the API.

    Args:
        device (VeSyncOutlet): The device object.
        details (ResponseDeviceDetailsModel): The device details.
        feature_map (OutletMap): The feature map for the device.

    Attributes:
        active_time (int): Active time of device, defaults to None.
        connection_status (str): Connection status of device.
        device (VeSyncOutlet): Device object.
        device_status (str): Device status.
        features (dict): Features of device.
        last_update_ts (int): Last update timestamp of device, defaults to None.
        energy (float): Energy usage in kWh.
        monthly_history (ResponseEnergyResult): Monthly energy history.
        nightlight_automode (str): Nightlight automode status.
        nightlight_brightness (int): Nightlight brightness level.
        nightlight_status (str): Nightlight status.
        power (float): Power usage in Watts.
        voltage (float): Voltage in Volts.
        weekly_history (ResponseEnergyResult): Weekly energy history.
        yearly_history (ResponseEnergyResult): Yearly energy history.

    Methods:
        update_ts: Update last update timestamp.
        to_dict: Dump state to JSON.
        to_json: Dump state to JSON string.
        to_jsonb: Dump state to JSON bytes.
        as_tuple: Convert state to tuple of (name, value) tuples.

    Note:
        Not all attributes are available on all devices. Some attributes may be None or
        not supported depending on the device type and features. The attributes are set
        based on the device features and the API response.
    """

    __slots__ = (
        'current',
        'currentUpperThreshold',
        'energy',
        'monthly_history',
        'nightlight_automode',
        'nightlight_brightness',
        'nightlight_status',
        'power',
        'protectionStatus',
        'voltage',
        'voltageUpperThreshold',
        'weekly_history',
        'yearly_history',
    )

    def __init__(
        self,
        device: VeSyncOutlet,
        details: ResponseDeviceDetailsModel,
        feature_map: OutletMap,
    ) -> None:
        """Initialize VeSync Switch State."""
        super().__init__(device, details, feature_map)
        self._exclude_serialization = [
            'weakly_history',
            'monthly_history',
            'yearly_history',
        ]
        self.device: VeSyncOutlet = device
        self.features: list[str] = feature_map.features
        self.active_time: int | None = 0
        self.power: float | None = None
        self.energy: float | None = None
        self.voltage: float | None = None
        self.current: float | None = None
        self.nightlight_status: str | None = None
        self.nightlight_brightness: int | None = None
        self.nightlight_automode: str | None = None
        self.weekly_history: ResponseEnergyResult | None = None
        self.monthly_history: ResponseEnergyResult | None = None
        self.yearly_history: ResponseEnergyResult | None = None
        self.voltageUpperThreshold: float | None = None
        self.protectionStatus: str | None = None
        self.currentUpperThreshold: float | None = None

    def annual_history_to_json(self) -> None | str:
        """Dump annual history."""
        if not self.device.supports_energy:
            logger.info('Device does not support energy monitoring.')
            return None
        if self.yearly_history is None:
            logger.info('No yearly history available, run device.get_yearly_history().')
            return None
        return self.yearly_history.to_json()

    def monthly_history_to_json(self) -> None | str:
        """Dump monthly history."""
        if not self.device.supports_energy:
            logger.info('Device does not support energy monitoring.')
            return None
        if self.monthly_history is None:
            logger.info('No monthly history available, run device.get_monthly_history().')
            return None
        return self.monthly_history.to_json()

    def weekly_history_to_json(self) -> None | str:
        """Dump weekly history."""
        if not self.device.supports_energy:
            logger.info('Device does not support energy monitoring.')
            return None
        if self.weekly_history is None:
            logger.info('No weekly history available, run device.get_weekly_history().')
            return None
        return self.weekly_history.to_json()


class VeSyncOutlet(VeSyncBaseToggleDevice):
    """Base class for Etekcity Outlets.

    State is stored in the `state` attribute of the device.
    This is only for holding state information and does not
    contain any methods for controlling the device or retrieving
    information from the API.

    Args:
        details (ResponseDeviceDetailsModel): The device details.
        manager (VeSync): The VeSync manager.
        feature_map (OutletMap): The feature map for the device.

    Attributes:
        state (OutletState): The state of the outlet.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
    """

    __slots__ = ('_energy_intervals', 'nightlight_modes')

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: OutletMap
    ) -> None:
        """Initialize VeSync Outlet base class."""
        super().__init__(details, manager, feature_map)
        self.state: OutletState = OutletState(self, details, feature_map)
        self._energy_intervals = feature_map.energy_intervals
        self.nightlight_modes = feature_map.nightlight_modes

    def _build_energy_request(self, method: str) -> RequestEnergyHistory:
        """Build energy request post."""
        request_keys = (
            'acceptLanguage',
            'accountID',
            'appVersion',
            'phoneBrand',
            'phoneOS',
            'timeZone',
            'token',
            'traceId',
            'userCountryCode',
            'debugMode',
            'homeTimeZone',
            'uuid',
        )

        body = Helpers.get_defaultvalues_attributes(request_keys)
        body.update(Helpers.get_manager_attributes(self.manager, request_keys))
        body.update(Helpers.get_device_attributes(self, request_keys))
        body['method'] = method
        return RequestEnergyHistory.from_dict(body)

    async def _get_energy_history(self, history_interval: str | EnergyIntervals) -> None:
        """Pull energy history from API.

        Args:
            history_interval (str): The interval for the energy history,
                options are 'getLastWeekEnergy', 'getLastMonthEnergy', 'getLastYearEnergy'

        Note:
            Builds the state.<history_interval>_history attribute.
        """
        if not self.supports_energy:
            logger.debug('Device does not support energy monitoring.')
            return
        if history_interval not in self._energy_intervals:
            logger.debug('Invalid history interval: %s', history_interval)
            return
        if not isinstance(history_interval, EnergyIntervals):
            history_interval = EnergyIntervals(history_interval)
        interval = ENERGY_HISTORY_MAP[history_interval]

        body = self._build_energy_request(interval)
        headers = Helpers.req_header_bypass()
        r_bytes, _ = await self.manager.async_call_api(
            f'/cloud/v1/device/{interval}',
            'post',
            headers=headers,
            json_object=body.to_dict(),
        )

        r = Helpers.process_dev_response(logger, history_interval, self, r_bytes)
        if r is None:
            return
        response = ResponseEnergyHistory.from_dict(r)
        match history_interval:
            case EnergyIntervals.WEEK:
                self.state.weekly_history = response.result
            case EnergyIntervals.MONTH:
                self.state.monthly_history = response.result
            case EnergyIntervals.YEAR:
                self.state.yearly_history = response.result

    @property
    def supports_nightlight(self) -> bool:
        """Return True if device supports nightlight.

        Returns:
            bool: True if device supports nightlight, False otherwise.
        """
        return OutletFeatures.NIGHTLIGHT in self.features

    @property
    def supports_energy(self) -> bool:
        """Return True if device supports energy.

        Returns:
            bool: True if device supports energy, False otherwise.
        """
        return OutletFeatures.ENERGY_MONITOR in self.features

    async def get_weekly_energy(self) -> None:
        """Build weekly energy history dictionary.

        The data is stored in the `device.state.weekly_history` attribute
        as a `ResponseEnergyResult` object.
        """
        await self._get_energy_history(EnergyIntervals.WEEK)

    async def get_monthly_energy(self) -> None:
        """Build Monthly Energy History Dictionary.

        The data is stored in the `device.state.monthly_history` attribute
        as a `ResponseEnergyResult` object.
        """
        await self._get_energy_history(EnergyIntervals.MONTH)

    async def get_yearly_energy(self) -> None:
        """Build Yearly Energy Dictionary.

        The data is stored in the `device.state.yearly_history` attribute
        as a `ResponseEnergyResult` object.
        """
        await self._get_energy_history(EnergyIntervals.YEAR)

    async def update_energy(self) -> None:
        """Build weekly, monthly and yearly dictionaries."""
        if self.supports_energy:
            await self.get_weekly_energy()
            await self.get_monthly_energy()
            await self.get_yearly_energy()

    async def set_nightlight_state(self, mode: str) -> bool:
        """Set nightlight mode.

        Available nightlight states are found in the `device.nightlight_modes` attribute.

        Args:
            mode (str): Nightlight mode to set.

        Returns:
            bool: True if nightlight mode set successfully, False otherwise.
        """
        del mode  # unused
        if not self.supports_nightlight:
            logger.debug('Device does not support nightlight.')
        else:
            logger.debug('Nightlight mode not configured for %s', self.device_name)
        return False

    async def turn_on_nightlight(self) -> bool:
        """Turn on nightlight if supported."""
        if not self.supports_nightlight:
            logger.debug('Device does not support nightlight.')
            return False
        return await self.set_nightlight_state(NightlightModes.ON)

    async def turn_off_nightlight(self) -> bool:
        """Turn off nightlight if supported."""
        if not self.supports_nightlight:
            logger.debug('Device does not support nightlight.')
            return False
        return await self.set_nightlight_state(NightlightModes.OFF)

    async def set_nightlight_auto(self) -> bool:
        """Set nightlight to auto mode."""
        if not self.supports_nightlight:
            logger.debug('Device does not support nightlight.')
            return False
        return await self.set_nightlight_state(NightlightModes.AUTO)
